/*
 * Modification Log:
 * -----------------
 * .01  02-07-94        bkc     Add the terse option
 * .02  04-12-94        bkc     Add the -n -w sec  options
 *				Allows get native waveform record
 * .03  03-29-95        bkc     Add the -f, -e options
 * .04  03-30-95        bkc     If enum string is null default to number
 *                              -f, -e works on integer type too
 * .05  mm-dd-yy        iii     Comment
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "chandata.h"

extern chandata *pchandata;
extern struct caGlobals CA;
extern union val_tag {
        char cv[16000];
        char pv[400][40];
        short sv[8000];
        int iv[4000];
        float fv[4000];
        double dv[2000];
	} wf_buf;

extern char *optarg;
extern int optind;
int TERSE=0,ENUM=0,FORMAT=0;
extern double atof();

main(argc,argv)
int argc;
char **argv;
{
char *str;
char f_fmt[15],e_fmt[15];
int c,status,value;
int count=0;

	CA.PEND_IO_TIME = 1.0;
if (argc < 2) {
	printf("\nUsage:  caget [-n] [-t] [-f n] [-e n] [-w sec] pv_name\n\n");
	printf("The only required input is the pv_name.\n");
	printf("caget calls channel access functions to return IOC string or numerical value\n");
	printf("to standard output for a requested process variable.\n\n");
	printf("        -n   returns numerical value instead of string for DBR_ENUM field\n\n");
	printf("	-t   Terse mode, process variable name not returned\n\n");
	printf("      -f n   Use f format, n specifies number of digits after the decimal point\n\n");
	printf("      -e n   Use e format, n specifies number of digits after the decimal point\n\n");
	printf("    -w sec   Wait time, specifies bigger time out, default is 1.0 second\n\n");
	printf("   pv_name   requested database process variable name\n\n");
	exit(1);
	}

while ((c = getopt(argc,argv,"ntf:e:w:")) != -1) 
	switch (c) {
	case 'n':
		ENUM = 1;
		break;
	case 't':
		TERSE = 1;
		break;
	case 'f':
		FORMAT = 1;
		sprintf(f_fmt,"%c-.%sf\n",37,optarg);
		break;
	case 'e':
		FORMAT = 2;
		sprintf(e_fmt,"%c-.%se\n",37,optarg);
		break;
	case 'w':
		CA.PEND_IO_TIME = atof(optarg);
		if (CA.PEND_IO_TIME < 1.0) CA.PEND_IO_TIME = 1.0;
		break;
	}

if (argv[optind] == NULL || strlen(argv[optind]) > NAME_LENGTH) {
	printf(" %s is an invlaid channel name\n", argv[optind]);
	exit(1);
	}

	ca_task_initialize();
	ca_find_dev(argv[optind],pchandata);
        if (pchandata->state == 3 || pchandata->state == 0) exit(1);
	count = ca_element_count(pchandata->chid);
	if (count > 1) {
		GetWaveForm2(argv[optind],count);
		}
	else {
	status = ca_get_dbr_string(pchandata); 
	if (pchandata->type != DBR_STRING) 
       		ca_get_conn_data(DBR_STS_DOUBLE,pchandata);

	switch (pchandata->type) {
	case DBR_STRING:
		if (!TERSE)
                printf("%-30s %s\n",argv[optind],pchandata->string);
		else 
                printf("%s\n",pchandata->string);
		break;

	case DBR_FLOAT:
	case DBR_DOUBLE:
		if (!TERSE) {
		if (FORMAT == 0)
			 printf("%-30s %g\n",argv[optind],pchandata->value);
		if (FORMAT == 1) {
			 printf("%-30s ",argv[optind]);
			 printf(f_fmt,pchandata->value);
			}
		if (FORMAT == 2) {
			 printf("%-30s ",argv[optind]);
			 printf(e_fmt,pchandata->value);
			}
		}
		else  {
		if (FORMAT == 0)
                	printf("%g\n",pchandata->value);
		if (FORMAT == 1)
                	printf(f_fmt,pchandata->value);
		if (FORMAT == 2)
                	printf(e_fmt,pchandata->value);
			}
               	break; 

	case DBR_ENUM:
		value = pchandata->value;
		if (ENUM == 1) {
		if (!TERSE) 
			printf("%-30s %d\n",argv[optind],value);
			else
                	printf("%d\n",value);
			}
		else if ( FORMAT == 0) {
		if (!TERSE) {
		if (strlen(pchandata->string) > 0)
                	printf("%-30s %s\n",argv[optind],pchandata->string);
		else 
                	printf("%-30s %d\n",argv[optind],value);
			}
		else {
		if (strlen(pchandata->string) > 0)
                	printf("%s\n",pchandata->string);
		else 
                	printf("%d\n",value);
			}
		}
		else {
		if (!TERSE) {
		if (FORMAT == 1)
			printf("%-30s ",argv[optind]);
                	printf(f_fmt,pchandata->value);
		if (FORMAT == 2)
			printf("%-30s ",argv[optind]);
                	printf(e_fmt,pchandata->value);
			}
		else {
		if (FORMAT == 1)
                	printf(f_fmt,pchandata->value);
		if (FORMAT == 2)
                	printf(e_fmt,pchandata->value);
			}
		}
		break;
	default:
		value = pchandata->value;
		if (!TERSE) {
		if (FORMAT == 0) 
                	printf("%-30s %ld\n",argv[optind],value);
		if (FORMAT == 1) {
			 printf("%-30s ",argv[optind]);
			 printf(f_fmt,pchandata->value);
			}
		if (FORMAT == 2) {
			 printf("%-30s ",argv[optind]);
			 printf(e_fmt,pchandata->value);
			}
		}
		else {
		if (FORMAT == 0)
                	printf("%ld\n",value);
		if (FORMAT == 1)
                	printf(f_fmt,pchandata->value);
		if (FORMAT == 2)
                	printf(e_fmt,pchandata->value);
		}

                }
	}
	ca_task_exit();

	exit(0);
}


/* get waveform record  as an double array */
int GetWaveForm(name,count)
char *name;
int count;
{
int i;
double *wf_value;
		wf_value = (double *)calloc(count,sizeof(double));
        	ca_get_wave_form(name,wf_value);
		if (!TERSE) printf("%s",name);
		printf(" %d ",count);
		for (i=0;i<count;i++) {
/*			if ( i%5 == 0) printf("\n");
*/
			printf("%g ",wf_value[i]);
			}	
		printf("\n");
		free(wf_value);
}


/* get native type waveform record */
int GetWaveForm2(name,no)
char *name;
int no;
{
int i;

        ca_get_native_wave_form(name,no,&wf_buf);
	if (!TERSE) printf("%s",name);
	printf(" %d ",no);
switch (pchandata->type) {
        case DBR_STRING:
        for (i=0;i<no;i++) {
		printf("%s ",wf_buf.pv[i]);
		}
        break;

        case DBR_SHORT:
        case DBR_ENUM:
	for (i=0;i<no;i++) {
                printf("%d ",wf_buf.sv[i]);
                }
        break;

/*      case DBR_UCHAR: */
        case DBR_CHAR:
	for (i=0;i<no;i++) {
                printf("%d ",wf_buf.cv[i]);
                }
        break;

/*      case DBR_USHORT: */
        case DBR_FLOAT:
        for (i=0;i<no;i++) {
                printf("%f ",wf_buf.fv[i]);
                }
        break;

/*      case DBR_ULONG: */
        case DBR_DOUBLE:
        for (i=0;i<no;i++) {
                printf("%f ",wf_buf.dv[i]);
                }
        break;

	default:
        case DBR_LONG:
	for (i=0;i<no;i++) {
                printf("%d ",wf_buf.iv[i]);
                }
        break;

	}
	printf("\n");

}
