/*
 * Modification Log:
 * -----------------
 * .01  02-07-94        bkc     Add the terse option
 * .02  02-21-95        bkc     Default to write as string
 * .03  03-17-95        bkc     Default to get/put as string except the
 *				DBR_ENUM type default to enumerate number
 * 				use -s to put enumerate string
 * .04  mm-dd-yy        iii     Comment
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "chandata.h"

extern chandata *pchandata;
extern struct caGlobals CA;

double atof();

extern char *optarg;
extern int optind;

main(argc,argv)
int argc;
char **argv;
{
int c,status,TERSE=0,ENUM_V=1,i;
char *str,string[MAX_STRING_SIZE];
struct dbr_gr_enum dbr;
double value;
CA.PEND_IO_TIME = 1.;

if (argc < 3) {
	printf("\nUsage:  caput [-t] [-s] [-w sec] pv_name  pv_value\n\n");
        printf("        -t   Terse mode, only the successfully put value is returned\n\n");
	printf("        -s   pv_value entered as database defined enumeral string\n\n");
        printf("    -w sec   Wait time, specifies bigger time out, default is 1 second\n\n");
        printf("   pv_name   requested database process variable name\n\n");
        printf("  pv_value   new value to put to IOC\n\n");

	exit(1);
	}

while ((c = getopt(argc,argv,"tsw:")) != -1) 
        switch (c) {
        case 't':
                TERSE = 1;
		break;
        case 's':
                ENUM_V = 0;
		break;
        case 'w':
                CA.PEND_IO_TIME = atof(optarg);
                if (CA.PEND_IO_TIME < 1.) CA.PEND_IO_TIME = 1.;
                break;
        }

if (argv[optind] == NULL || strlen(argv[optind]) > NAME_LENGTH) {
	printf(" %s is an invlaid channel name\n", argv[optind]);
	exit(1);
	}

	ca_task_initialize();

	ca_find_dev(argv[optind],pchandata);
	if (pchandata->state == cs_closed || pchandata->state == cs_never_conn)
		 exit(1);

/* get old value */
	str = (char *)ca_get_string(argv[optind]);
	if (!TERSE)
	printf("Old : %-30s %s\n",argv[optind],str);

/*  check for enum type */

	if (ENUM_V == 1 && pchandata->type == DBR_ENUM) {
		status = ca_get(DBR_GR_ENUM,pchandata->chid,&dbr);
		ca_pend_io(CA.PEND_IO_TIME);
		if (atoi(argv[optind+1]) >=  dbr.no_str) {
			fprintf(stderr,"Error:  Invalid value entered!\n");
			exit(1);
			}
		}

/* put value to device default as string except DBR_ENUM */

	value = atof(argv[optind+1]);
	strcpy(string,argv[optind+1]);
	if (argc > optind+1 ) {
	for (i=optind+2;i<argc;i++) {
		strcat(string," ");
		strcat(string,argv[i]); 
		}
	}

	if (ENUM_V == 1 && pchandata->type == DBR_ENUM) 
	status = ca_put_conn_data(DBR_DOUBLE,pchandata,&value);
	else
	status = ca_put_conn_data(DBR_STRING,pchandata,string);	

/*get new value */

	if (status == ECA_NORMAL) { 
	str = (char *)ca_get_string(argv[optind]);
	if (!TERSE)
		printf("New : %-30s %s\n",argv[optind],str);
		else
		printf("%s\n",str);
		}
	else 
		ca_check_return_code(status);
	ca_task_exit();

	exit(0);
}

